#include "main.h"

#include <wayland-client.h>
#include <wayland-egl.h>
#include <EGL/egl.h>
#include <GLES2/gl2.h>

//////////////////////////////////////////////////////////////////////////////

typedef struct
{
   struct wl_surface       *surf;
   struct wl_shell_surface *shell_surf;
   struct wl_egl_window    *egl_win;
   EGLContext               egl_ctx;
   EGLSurface               egl_surf;
   int                      w, h;
} Win;

//////////////////////////////////////////////////////////////////////////////

static struct wl_display    *disp       = NULL;
static struct wl_compositor *comp       = NULL;
static struct wl_shell      *shell      = NULL;
static EGLDisplay            egl_disp   = 0;
static Win                  *win        = NULL;

//////////////////////////////////////////////////////////////////////////////

static void reg_obj_add            (void *data, struct wl_registry *reg, uint32_t name, const char *interface, uint32_t version);
static void reg_obj_del            (void *data, struct wl_registry *reg, uint32_t name);
static void shell_surf_ping        (void *data, struct wl_shell_surface *shell_surf, uint32_t serial);
static void shell_surf_configure   (void *data, struct wl_shell_surface *shell_surf, uint32_t edges, int32_t w, int32_t h);
static void shell_surf_popup_done  (void *data, struct wl_shell_surface *shell_surf);
static void win_add                (Win *win, int w, int h);
static void win_del                (Win *win);
static void win_render             (Win *win);

static struct wl_registry_listener reg_listener =
{
   reg_obj_add,
   reg_obj_del
};

static struct wl_shell_surface_listener shell_surf_listener =
{
   shell_surf_ping,
   shell_surf_configure,
   shell_surf_popup_done
};

static void
reg_obj_add(void *data, struct wl_registry *reg, uint32_t name,
            const char *interface, uint32_t version)
{
   if (!strcmp(interface,"wl_compositor"))
     {
        comp = wl_registry_bind(reg, name, &wl_compositor_interface, 1);
     }
   else if (!strcmp(interface,"wl_shell"))
     {
        shell = wl_registry_bind(reg, name, &wl_shell_interface, 1);
     }
}

static void
reg_obj_del(void *data, struct wl_registry *reg, uint32_t name)
{
}

static void
shell_surf_ping(void *data, struct wl_shell_surface *shell_surf,
                uint32_t serial)
{
   wl_shell_surface_pong(shell_surf, serial);
}

static void
shell_surf_configure(void *data, struct wl_shell_surface *shell_surf,
                     uint32_t edges, int32_t w, int32_t h)
{
   Win *win = data;
   wl_egl_window_resize(win->egl_win, w, h, 0, 0);
}

static void
shell_surf_popup_done(void *data, struct wl_shell_surface *shell_surf)
{
}

//////////////////////////////////////////////////////////////////////////////

static void
win_add(Win *win, int w, int h)
{
   EGLConfig cfg;
   EGLint num_cfg;
   EGLint attr[] =
     {
        EGL_SURFACE_TYPE    , EGL_WINDOW_BIT,
        EGL_RENDERABLE_TYPE , EGL_OPENGL_ES2_BIT,
        EGL_RED_SIZE        , 8,
        EGL_GREEN_SIZE      , 8,
        EGL_BLUE_SIZE       , 8,
        EGL_ALPHA_SIZE      , 8,
        EGL_DEPTH_SIZE      , 24,
        EGL_STENCIL_SIZE    , 8,
        EGL_NONE
     };
   EGLint attr2[] =
     {
        EGL_CONTEXT_CLIENT_VERSION, 2,
        EGL_NONE
     };

   eglBindAPI(EGL_OPENGL_ES_API);
   eglChooseConfig(egl_disp, attr, &cfg, 1, &num_cfg);
   win->egl_ctx = eglCreateContext(egl_disp, cfg, EGL_NO_CONTEXT, attr2);

   win->surf = wl_compositor_create_surface(comp);
   win->shell_surf = wl_shell_get_shell_surface(shell, win->surf);
   wl_shell_surface_add_listener(win->shell_surf, &shell_surf_listener, win);
   wl_shell_surface_set_toplevel(win->shell_surf);
   win->egl_win = wl_egl_window_create(win->surf, w, h);
   win->w = w;
   win->h = h;
   win->egl_surf = eglCreateWindowSurface(egl_disp, cfg, win->egl_win, NULL);
   eglMakeCurrent(egl_disp, win->egl_surf, win->egl_surf, win->egl_ctx);
}

static void
win_del(Win *win)
{
   eglDestroySurface(egl_disp, win->egl_surf);
   wl_egl_window_destroy(win->egl_win);
   wl_shell_surface_destroy(win->shell_surf);
   wl_surface_destroy(win->surf);
   eglDestroyContext(egl_disp, win->egl_ctx);
}

//////////////////////////////////////////////////////////////////////////////

static void
win_render(Win *win)
{
   static int init = 0;

   if (!init)
     {
        init = 1;
        draw_init(win->w, win->h);
     }
   draw(win->w, win->h);
   eglSwapBuffers(egl_disp, win->egl_surf);
}

//////////////////////////////////////////////////////////////////////////////

int
main(int argc, char **argv)
{
   win = calloc(1, sizeof(*win));
   disp = wl_display_connect(NULL);

   struct wl_registry *reg = wl_display_get_registry(disp);
   wl_registry_add_listener(reg, &reg_listener, NULL);
   wl_display_roundtrip(disp);

   egl_disp = eglGetDisplay(disp);
   eglInitialize(egl_disp, NULL, NULL);

   win_add(win, 1920, 1080);

   for (;;)
     {
        wl_display_dispatch_pending(disp);
        win_render(win);
     }

   win_del(win);
   eglTerminate(egl_disp);
   wl_display_disconnect(disp);

   return 0;
}
